#!/usr/bin/python

import os, sys
import util
import pprint
import random
import configparser
from optparse import OptionParser
from macealtree.sentences import Sentences
from pathlib import Path

# eval individual annotators against gold standard
# and also the best prediction made by MACE
def evalAnnotations(goldfile, sentences, tagdict_file):

        # read the gold standard 
        gold = readGold(goldfile, sentences, tagdict_file)
        
        # evaluate each annotator against the goldstandard
        result = sentences.eval_labels(gold)

        # and also add the results for a majority vote
        result += sentences.majority_baseline(gold)
               
        return result 

        

# read the goldstandard file and return it as a list
def readGold(gold, sentences, tagdict_file):
        
        goldlist = []
        sentdict = {}
        sid = 0
        # read gold file
        gold = util.load_list_from_file(gold)
        for line in gold:
                if line == '\n' or line == '':
                    sid += 1
                else:
                        tid, wrd, lem, tag, x, y, head, dep, a, b = line.split("\t") 
                        tid = int(tid) -1
                        head = int(head) -1
                        # new, unknown label? add to dictionary
                        if dep not in sentences.tagdict:
                                newTag = len(sentences.tagdict) +1
                                sentences.tagdict[dep] = newTag
                                util.write_dict_keys_to_file(sentences.tagdict, tagdict_file)
                        goldlist.append(str(sid)+":"+str(tid)+":"+str(head)+":"+dep)
        return goldlist
    

                        
# convert table content to numerical representation
# e.g. dobj = 8, case = 9, ...
def convert2num(table, sentences):
    # convert depencency labels to numbers
    for x in range(0, len(table['l'])):
            for y in range(0, len(table['l'][x])):
                lab = table['l'][x][y]  
                table['l'][x][y] = sentences.tagdict[lab]
    return table

 


def run_iteration(options, sentences, table, tagdict_file, c):
        #####
        # create new annotator from best predictions and get N tags for the predictions 
        # with the highest entropy and update the new annotator (feedback from oracle):
        # replace a randomly chosen annotator in every round
        bad_guy = random.choice(range(len(table['arc'])))
        
        # select table entry with highest entropy
        # returns a list with row indices for the rows
        # in the table that have the highest entropy (from file "entropy" created by MACE)
        N = 1
        indices = sentences.entropy(N) 
        print
        for ind in indices:
                if options.simulation:
                        # either from the gold standard (AL simulation)
                        sid, tid, tag, head = sentences.getOracleTag(ind)
                else:
                        # or from the oracle (real AL with human oracle)
                        sid, tid, tag, head = sentences.getAnnotation(ind)

                if tag == "now" and head == "stop":
                        return
                else:
                	print("NO STOP " + tag + " " + head)
                sid += 1; 
                tid += 1; 
                head = int(head) + 1
                util.print_log(logfile, "Sentence " + str(sid) + ",\ttoken " + str(tid) + " \tTAG: " + tag + "\tHEAD: " + str(head) + "\n")

        # now use oracle tags as feedback, combined with the predlist features from MACE
        # create control file and rerun EM
        if options.feedback:
            if c%2 == 1: 
                sentences.write_feedback_to_file("feedback.attach", bad_guy, 1)
                sentences.update_predictions(bad_guy, 1)
            else:
                sentences.write_feedback_to_file("feedback.deprel", bad_guy, 0)
                sentences.update_predictions(bad_guy, 0) 
        sentences.backupFiles(c)


        # run mace on labels or on attachments (take turns!)
        if c%2 == 1:
            # run mace on attachments
            os.system("cp preds-attach.csv preds.csv")
            os.system("cp prediction prediction.deprel")
        else:
            # run mace on deprel labels
            os.system("cp preds-deprel.csv preds.csv")
            os.system("cp prediction prediction.attach")

        # run MACE 
        if options.entropies:
                if options.feedback:
                        command = "java -jar ./MACE.jar --controls feedback --entropies preds.csv"
                else:
                        command = "java -jar ./MACE.jar --entropies preds.csv"
        else:
                if options.feedback: 
                        command = "java -jar ./MACE.jar --controls feedback preds.csv"
                else:
                        command = "java -jar ./MACE.jar preds.csv"
        if options.restarts:
                newarg = "MACE.jar --restarts " + options.restarts
                command = command.replace("MACE.jar", newarg)
        if options.vanilla:
            newarg = "MACE.jar --em "
            command = command.replace("MACE.jar", newarg)

        os.system(command)

        # read MACE predictions
        deprel = Path('prediction.deprel')
        attach = Path('prediction.attach')
        if deprel.is_file() and attach.is_file():
            predlist_deprel = util.load_list_from_file('prediction.deprel')
            predlist_attach = util.load_list_from_file('prediction.attach')

            # done? now print the trees 
            if config['MACE-AL-TREE']['PRINT_TREE'] == 'CLE':
                util.print_trees_CLE(predlist_deprel, predlist_attach, sentences, c)
            elif config['MACE-AL-TREE']['PRINT_TREE'] == 'CLE_weighted':
                util.print_trees_CLE_weighted(predlist_deprel, predlist_attach, sentences, c)
            else:
                util.print_trees(predlist_deprel, predlist_attach, sentences, c)

            util.print_trees_CLE_weighted(predlist_deprel, predlist_attach, sentences, c) 




###########################################################################
###
###   main
###
###########################################################################

        
if __name__ == '__main__':

#def run_cmd_interface():

        # parse command line options
        optparser = OptionParser()
        optparser.add_option("-f", "--feedback", dest="feedback", default=False,
                             help="read feedback from oracle annotations (optional)")
        optparser.add_option("-e", "--entropies", dest="entropies", default=False,
                             help="print entropies for each instance to file (MACE)")
        optparser.add_option("-m", "--majority", dest="majority", default=False,
                             help="instead of posterior entropies, use majority vote")
        optparser.add_option("-r", "--restarts", dest="restarts", default=False,
                             help="number of random model initialisations")
        optparser.add_option("-s", "--simulation", dest="simulation", default=False,
                             help="run program as AL simulation")
        optparser.add_option("-v", "--vanilla", dest="vanilla", default=False,
                            help="run program with vanilla EM")
        optparser.add_option("-l", "--lowest", dest="lowest", default=False,
                             help="replace annotator with the lowest competence")

        (options, args) = optparser.parse_args() 
        
	# read in config file
        config = configparser.ConfigParser()
        config.read('config/mace-al-tree.config')
        logfile = config['MACE-AL-TREE']['LOGFILE']
        datadir = config['MACE-AL-TREE']['DATA_DIR']
        goldstandard = 'False'
        if config['MACE-AL-TREE']['GOLD_FILE']:
	        goldstandard = datadir + "/" + config['MACE-AL-TREE']['GOLD_FILE']
        
        tagdict_file = "dict/tags.txt"
        if config['MACE-AL-TREE']['TAG_DICT']:
                tagdict_file = config['MACE-AL-TREE']['TAG_DICT']
        tagdict = util.load_dict_from_file(tagdict_file)

        # read gold standard (or, if no gold annotations exist, read list with word tokens)
        sentences = Sentences(goldstandard, tagdict)
        
        # read predictions from file
        table = util.readPredictions(datadir, config['MACE-AL-TREE']['PRED_SUFFIX'])
       
        # add predictions
        sentences.addPredictions(table) 

        # convert tags to numerical representation
        table = convert2num(table, sentences) 

        # print dict content to csv file
        sentences.printCSV()


        ########
        # before we start with error correction, let's evaluate the individual annotators
        # and get the majority vote baseline (only if there is a goldstandard available)
        if goldstandard != False:
                util.print_log(logfile, evalAnnotations(goldstandard, sentences, tagdict_file))

        
        # run MACE
        if options.entropies:
                command = "java -jar ./MACE.jar --entropies preds-deprel.csv"
        else:
                command = "java -jar ./MACE.jar preds-deprel.csv"
        if options.restarts:
                newarg = "MACE.jar --restarts " + options.restarts
                command = command.replace("MACE.jar", newarg)
        if options.vanilla:
                newarg = "MACE.jar --em "
                command = command.replace("MACE.jar", newarg)

        util.print_log(logfile, 'Running MACE:\n')
        util.print_log(logfile, command+'\n')
        util.print_log(logfile, '\nSTART ERROR CORRECTION:\n\n')

        # before running mace, convert preds.csv to numerical format
        # we need two tag dicts: one for the labels, the other one for the attachments
        # then we alternate running MACE on the attachments and the labels

        # run MACE on deprel 
        os.system(command)  
        # back up predictions for deprel
        os.system("cp prediction prediction.deprel") 
        # run MACE on attachments (replace deprel by attach in command)
        command = command.replace("deprel", "attach");
        os.system(command)  
        # backup deprel file and run MACE on attachments 
        os.system("cp prediction prediction.attach")

        # read MACE predictions
        predlist_deprel = util.load_list_from_file('prediction.deprel')
        predlist_attach = util.load_list_from_file('prediction.attach')

        # and print trees for evaluation 
        if config['MACE-AL-TREE']['PRINT_TREE'] == 'CLE':
	        util.print_trees_CLE(predlist_deprel, predlist_attach, sentences, 0)
        elif config['MACE-AL-TREE']['PRINT_TREE'] == 'CLE_weighted':
                util.print_trees_CLE_weighted(predlist_deprel, predlist_attach, sentences, 0) 
        else:
                util.print_trees(predlist_deprel, predlist_attach, sentences, 0)

        # run c iterations of AL
        for c in range(0, int( config['MACE-AL-TREE']['NO_ITERATIONS'] )):
            run_iteration(options, sentences, table, tagdict_file, c)

        # read MACE predictions
        predlist_deprel = util.load_list_from_file('prediction.deprel')
        predlist_attach = util.load_list_from_file('prediction.attach')

        # done? now print the trees
        treefile = ""
        if config['MACE-AL-TREE']['PRINT_TREE'] == 'CLE':
                util.print_trees_CLE(predlist_deprel, predlist_attach, sentences, c)
                firsttree = "trees1-cle.txt"
                treefile = "trees" + str(c) + "-cle.txt"
        elif config['MACE-AL-TREE']['PRINT_TREE'] == 'CLE_weighted':
                util.print_trees_CLE_weighted(predlist_deprel, predlist_attach, sentences, c)
                firsttree = "trees1-cle-w.txt"
                treefile = "trees" + str(c) + "-cle-w.txt"
        else:
                util.print_trees(predlist_deprel, predlist_attach, sentences, c)
                firsttree = "trees1.txt"
                treefile = "trees" + str(c) + ".txt"


        # and evaluate the final tree against the gold standard
        util.print_log(logfile, "\n\nEvaluation after 1 iteration of error correction:")
        command = "perl eval07.pl -m -q -g " + goldstandard + " -s " + firsttree + " >>" + logfile
        os.system(command)

        util.print_log(logfile, "\nEvaluation after " + str(c+1) + " iterations of error correction:")
        command = "perl eval07.pl -m -q -g " + goldstandard + " -s " + treefile + " >>" + logfile
        os.system(command)
        
        if config['MACE-AL-TREE']['DELETE_TEMP'] == 'True':
        	command = "mv " + treefile + " final_trees.conll"
        	os.system(command)
        	command = "bash clean.sh"
        	os.system(command)
